import socket

from plogger import logger


class SocketClient:
    """Create socket and establish connect to service using tuple host+port"""

    greeting = None

    def __init__(self,
                 host: str,
                 port: int = 0,
                 buffer_size: int = 4096,
                 initialize: bool = False,
                 logger_enabled: bool = True,
                 log_level: str = 'INFO',
                 connection_timeout: int = None):
        """Create and connect client to a remote host.

        :param host: Host IP
        :param port: Port
        :param logger_enabled: Enable/disable module logger
        :param log_level: INFO
        :param initialize: Establish connection during init
        """

        self.host = host
        self.port = port
        self.buffer_size = buffer_size
        self.logger = logger('SocketClient', enabled=logger_enabled, level=log_level)
        self.connection_timeout = connection_timeout

        if initialize:
            try:
                self.client = self.connect(timeout=self.connection_timeout)
            except ConnectionRefusedError as err:
                self.logger.error(f'Cannot establish socket connection to {self.host}:{self.port}. {err}')
                raise err
            except socket.gaierror as err:
                self.logger.error(f'Check host and port format. {self.host}:{self.port}. {err}')
                raise err
            except socket.timeout as err:
                self.logger.error(f'{self.host}:{self.port} is unavailable within {self.connection_timeout} sec. {err}')
                raise err

    def __getattr__(self, attr):
        self.logger.error(f'No such attribute ({attr}) error. Perhaps, object is not initialized.')
        raise AttributeError(f'SocketClient does not have specific attribute "{attr}"')

    def __str__(self):
        msg = (f'Host: {self.host}\n'
               f'Port: {self.port}\n'
               f'Reading buffer size: {self.buffer_size}\n'
               f'Logger name: {self.logger.name}\n'
               f'Connection timeout: {self.connection_timeout}')
        return msg

    def connect(self, timeout: int = None):
        """Create connection

        :param timeout:
        :return:
        """

        self.logger.debug(f'Connecting to the <{self.host}:{self.port}>')
        conn = socket.create_connection((self.host, self.port), timeout=timeout)
        self.logger.debug('Connection has been established')

        return conn

    def is_socket_available(self,
                            port: int = 0,
                            host: str = None,
                            timeout: int = 5,
                            logger_enabled: bool = True) -> bool:
        """Check remote socket is available.

        Port 0 used by default. Used port from construct is not specified.

        :param host:
        :param port:
        :param timeout:
        :param logger_enabled:
        """

        host_ = host if host else self.host
        port_ = port if port else self.port

        try:
            with socket.create_connection((host_, port_), timeout=timeout) as sock:
                sock.settimeout(None)
                if logger_enabled:
                    self.logger.info(f'{host_}:{port_} | available')
                return True
        except socket.timeout:
            if logger_enabled:
                self.logger.info(f'{host_}:{port_} | unavailable')
            return False

    def wait_socket_available(self, port: int = 0, host: str = None, timeout: int = 5):
        """Wait for a socket availability

        :param port:
        :param host:
        :param timeout:
        :return:
        """

        timer = 1
        status = self.is_socket_available(port=port, host=host, timeout=1, logger_enabled=False)

        while not status:
            status = self.is_socket_available(port=port, host=host, timeout=1, logger_enabled=False)
            timer += 1

            if timer > timeout:
                error_msg = f'The service was not started within {timeout} seconds.'
                self.logger.error(error_msg)
                raise TimeoutError(error_msg)
        return status

    def send_command(self, cmd: str = ''):
        """Send network command and receive response.

        :param cmd:
        """

        command = self._encode_command(cmd)

        self.logger.info(f'{self.host:<14} | -> <{cmd}>')

        try:
            self.client.sendall(command)

            response = self.receive_all()

            # Save the greeting message for the first time only
            if self.greeting is None:
                self.logger.debug(f'Store "greetings": {response}')
                self.greeting = response

            return response
        except AttributeError as err:
            self.logger.error(f'{self.host:<14} | {err}')
            raise err

    def receive_all(self, stop_signs: str = '\r\n\r\n'):
        """Get and decode socket response

        Args:
            buffer_size: 4096 by default
            stop_signs: Symbols which will stop socket reading
        """

        self.logger.debug(f'Start reading from socket.\n\t'
                          f'{self.buffer_size = }\n\t'
                          f'{stop_signs = }')

        data = bytearray()

        self.logger.debug(f'Initial data: {data}')

        while True:
            chunk = self.client.recv(self.buffer_size)
            self.logger.debug(f'Chunk ({len(chunk)}): {chunk}')

            data.extend(chunk)

            if data.endswith(stop_signs.encode()):
                self.logger.debug(f'Reading response ran to completion. <{stop_signs = }> received.')
                break

        self.logger.debug('Parse response: decode, strip and splitlines')
        response = data.decode().strip().splitlines()
        self.logger.debug('Completed')

        self.logger.info(f'{self.host:<14} | <- {response}')
        return response

    def close_connection(self):
        self.client.close()

    def _encode_command(self, cmd):
        """Encode command to send"""

        self.logger.debug(r'Add "\n" symbol and encode command')
        return (cmd + '\n').encode()

    def get_sock_name(self) -> tuple:
        """Get local IP and port"""

        return self.client.getsockname()

    def get_peer_name(self) -> tuple:
        """Get remote IP and port"""
        return self.client.getpeername()
